(* Copyright (c) 2002 INRIA *)

(**  {b This module gives a description of the various datatypes
   and signatures used in the implementation of the server }
*)

(** {b Basic Structures} *)

(** 
   [method_name] is the type that encloses the name of the
   different methods present in the server.
*)
type method_name = [ Meth_Name of string ];

(** Each method is stored in a table as a list of entries 
   encapsulated as [table_entry]. Each entry contains a 
   [method_name] and an OCaml function which takes a list
   of parameters of type [SoapTypes.ty] and returns a value
   of type [SoapTypes.ty] .
*)
type table_entry = [Entry of method_name and (list SoapTypes.ty -> SoapTypes.ty)];


(** {b Server Module } *)

(** 
   This minimal XmlRPC Server is the functional part of the 
   main server. The module [SERVER] consists of its database,
   i.e., its list of functions. [table] stores this as a 
   list of [table_entry].
*)
(**  The methods [add_handler], [remove_handler] and [execute] 
   are the three  basic operations that are permitted by this 
   server. Other operations like [activate_method] etc. , 
   may be added later.
*)
(** Exception [Handler_Not_Present] is raised, if a request
   is obtained to remove, or execute a method, that does not
   exist in the server.
   
*)
(** Exception [Handler_Already_Present] is raised, if an attempt is made to add
a handler for a method that is already present. *)

module type SERVER = 
sig
     value table : ref (list table_entry) ; 
     (** Database of the handlers present in the server *)

     value add_handler : method_name -> (list SoapTypes.ty -> SoapTypes.ty) -> unit ;
     (** Adds the handler into the database. Raises an exception if the method is
     already present*)

     (** Removes the handler from the database. Raises an exception if the
     method is not present *)
     value remove_handler : method_name -> unit ;

     value execute : method_name -> list SoapTypes.ty  -> SoapTypes.ty ;
     (** Executes the given method and returns the result *)

     (** Raised if an attempt is made to add a handler for a method that is
     already present *)
     exception Handler_Already_Present; 

     exception Handler_Not_Present;
     (** Raised if an attempt is made to evaluate or remove a method that does
     not exist in the server *)

end ;

module Server : SERVER;


(** {b HTTP Server Module} *)
(** 
   The module [Webserver] has been designed as a functor,
   that takes a module of type [SERVER] and returns an
   http server, that can handle http requests. 
*)
(** The function [establish_connection] expects the server
   name and the port number as arguments in the form of
   strings. 
*)
(** Once the connection is established, it reads the message
   and extracts the {i XML Document }. This document, which is in
   the form of a string, is used by the function 
   [handle_msg], which processes this string. It then returns
   the processed document, and sends the document back to the 
   client.
*)
(** Exceptions like [Read_Error] and [Write_Error] are raised
   when there is some problem with the network, or some other 
   I/O error etc.
   
*)

module Webserver : 
    functor (Sr : SERVER) ->
        sig
          (** Raised if there is some {i read error} from some read operation
          from a {i file descriptor} *)
          exception Read_Error ; 

          (** Raised if there is some {i write error} from some write operation
          to {i file descriptor} *)
          exception Write_Error ; 

          (** Establishes connection at the specified address and port. It then
              listens for HTTP requests *)
          value establish_connection : string -> string -> unit ;

          (** Receives the message from the client in the form of a string. It
          then processes that string and replies back to the HTTP Client *)
          value handle_msg : string -> string ;
        end ;
