(* Copyright (c) 2002 INRIA *)

open Netencoding;
open Http_client;
open SoapTypes;
open Pxp_yacc;
open Pxp_document;
open Pxp_types;
open Format;

exception Request_failed ;
exception Request_fault of SoapTypes.soapst;

value xml_version = "<?xml version=\"1.0\"?>";
value client_version = "0.0.1";

value agent="Ocaml-SOAP/" ^ client_version ^ " (" ^ Sys.os_type ^ ")";

value format_params buf args = format_one args 
  where rec format_one ls =
    match ls with
      [ [car :: cdr] -> let _ = do {
	(*  Buffer.add_string buf "<param>\n";*)
	  Buffer.add_string buf (SoapTypes.print_value car);
	  (*Buffer.add_string buf "</param>\n"*)} in
	  format_one cdr
      | [] -> ()
      ] ;

value format_request b f args = 
  let buf = Buffer.create 512 in
  let _ = if (True) then do {
    Buffer.add_string buf 
      ("<Envelope><Body><methodName>"^ f ^ "</methodName>");
    Buffer.add_string buf "<value>";
    format_params buf args;
    Buffer.add_string buf "</value></Body></Envelope>"} 
    else do {
             Buffer.add_string buf
                   (xml_version ^ "\n<methodRem>\n<methodName>" ^ f ^"</methodName>\n");
             Buffer.add_string buf "</methodRem>\n"}
                   
    in
    Buffer.contents buf;
      
value parse_params ls = match ls with 
  [ [p::[]] -> SoapTypes.parse_value p
  | _ -> raise (Request_failed (*"Malformed response"*))
  ];

value nodename n = match n#node_type with
  [ T_element name -> name
  | _ -> "D"
  ];

value handle_response xml =
  try
    let methresp =
      parse_wfdocument_entity default_config (from_string xml) default_spec in
    let repval = List.hd methresp#root#sub_nodes in
    let repname =nodename repval in
    let body = repval#sub_nodes in
      if repname = "fault" then
	let fault_struct = SoapTypes.parse_value (List.hd body) in
	raise (*(Request_fault (SoapTypes.ml_of_struct fault_struct))*)Request_failed 
      else (* params *)
	parse_params (List.hd body)#sub_nodes
  with
    [ Validation_error e ->
	raise (Request_failed (*("Malformed response: " ^ e)*))
    | WF_error e ->
	raise (Request_failed (*("Malformed response: " ^ e)*))
    | At (wh, e) -> 
	let msg = match e with
	  [ Validation_error e -> e
          | WF_error e -> e
	  | _ -> "PXP error" 
          ] in
	raise (Request_failed (*("Malformed response: " ^ msg)*))
    ];
    
value handler = new pipeline;
handler#set_proxy_from_environment();
value options = handler#get_options ;

value _ = 
        handler#set_options { (options) with
                verbose_connection = True ;
                verbose_request_header = True ;
                verbose_request_contents = True;
                verbose_response_header = True ;
                verbose_response_contents = True;
                verbose_status = True 
        };

value make_request m trials = 
  do {
  handler#add m;
  let rec next_trial todo e =
    if todo > 0 then 
      try
	handler#run()
      with
      [ Http_client.Http_error(n,s) as e' -> 
	    if List.mem n [408; 413; 500; 502; 503; 504 ] then
	      next_trial (todo-1) e'
	    else
	      raise e'
	| e' -> next_trial (todo-1) e'
      ]
    else
      raise e
  in
    next_trial trials (Failure "bad number of http_trials")
    } ;

value do_request b server f args =
  let methodcall = format_request b f args in
  let req = new Http_client.post_raw server methodcall in
  let _ = req#set_req_header "Connection: close" in
  let _ = req#set_req_header "Content-Type" "text/xml" in
  let _ = req#set_req_header "User-Agent" agent in
  let _ = req#set_req_header "Content-length" 
      (string_of_int (String.length methodcall)) in
  let _ =  make_request req 5 in
    if not req#is_served then 
      raise (Request_failed (*"Couldn't get valid response from server"*))
    else
      (*handle_response (req#get_resp_body());*)Unit;

class remote host func =
object
  method zero_call () = do_request True host func [];
  method simple_call arg = do_request True host func [arg];
  method call = do_request True host func;
  method remove () = do_request False host func [];
end;
    
value  r = new remote "http://quincy.inria.fr:2816" "double" ;
value par = new Decimal_s.decimal_s ;
value _ = (par#set_struct) "dec" SoapXS.dec_sch ;
value _ = (par#set_obj) "<dec>2.0</dec>" ;
r#simple_call (sp_of_decimal par);
