(* Copyright (c) 2002 INRIA *)

open Pxp_yacc ;
open Pxp_document ;
open Pxp_types ;
open SoapTypes ;

value soap_env_prefix = "SOAP-ENV" ;
value soap_env_uri = "http://schemas.xmlsoap.org/soap/envelope/";

value schema_xsi_prefix = "xsi" ;
value schema_xsi_uri = "http://www.w3.org/2001/XMLSchema-instance";

value schema_xsd_prefix = "xsd" ;
value schema_xsd_uri = "http://www.w3.org/2001/XMLSchema";

type qname = [QName of string] ;
type table = (Hashtbl.t qname string) ;

exception Bad_argument ;
exception No_header_present ;
exception No_header_entry_present ;
exception No_body_entry_present ; 
exception No_Type_info ;

value nodename n = match n#node_type with
   [ T_element name -> name
   | T_attribute attr -> attr
   | _ -> "D"
   ];

value rec zip f ls = match ls with
     [ [] -> []
     | [x::l] -> [(f x) :: (zip f l)]
     ];          

value myAnd b1 b2 = not(or (not b1) (not b2)) ;

value default = "<def></def>" ;
value def =  parse_wfdocument_entity default_config (from_string default) default_spec ;
value def = def#root ;

class attributehandler = 
        object (self)
                value attributes:table  = Hashtbl.create 1 ;

                value namespace_list = ref [] ;

                method set_attr q str = match (q,str) with
                                     [ (_,"") -> raise Bad_argument
                                     | (QName n,v) -> Hashtbl.replace attributes (QName n) v 
                                     ];

                method get_attr q = match q with
                                     [ QName n -> Hashtbl.find attributes (QName n) ];

                method rem_attr q = match q with
                    [ QName n -> Hashtbl.remove attributes (QName n)];

                method declareNS prefix uri = match (prefix,uri) with
                                     [ ("","") -> () 
                                     | (p,u) -> do {namespace_list.val :=
                                             namespace_list.val @ [(p,u)] }
                                     ]; 

                method private print_ns (p,u) = String.concat "" ["xmlns:";p;u;"\n"] ;

                method ns_to_string () = 
                        String.concat "" (zip self#print_ns (\!(namespace_list))) ;

                value mutable str = "" ;
                method private print_att n v = match n with
                     [ QName m ->
                        str := String.concat "" [str;m;"=";v;"\n"]
                     ];

                method att_to_string () = 
                        let _ = Hashtbl.iter (self#print_att) attributes in
                           str ;
                        
                method to_string () = "B" ;(* To be defined later *)
        end ;


class header_entry =
        object (self)
                
                value mutable entry_name = "" ;
                
                value mutable v = sp_of_struct(make_struct());

                value mutable sch = def ;
                
                value attr_handl = new attributehandler ;
                
                method setname n = entry_name := n ;

                method getname () = entry_name ;

                method setval vv = v := vv ;

                method getval () = v ; 

                method set_sch (s:Pxp_document.node 'a) = sch := s ;

                method get_sch () = sch ;

                method set_attr q str = match (q,str) with
                                    [ (_,"") -> raise Bad_argument
                                    | (QName n,v) -> attr_handl#set_attr (QName n) v 
                                    ];

                method get_attr q = match q with 
                   [ QName n -> attr_handl#get_attr (QName n)] ;

                method rem_attr q = match q with
                   [QName n -> attr_handl#rem_attr (QName n)] ;

                method declareNS prefix uri = match (prefix,uri) with
                                    [ ("","") -> ()
                                    | (p,u) -> attr_handl#declareNS p u 
                                    ];

                method to_string () = 
                        let s1 = String.concat "" ["<";entry_name] in
                        let s2 = attr_handl#ns_to_string() in
                        let s3 = attr_handl#att_to_string() in
                        let s4 = SoapTypes.print_value v in
                        let s5 = String.concat "" ["</";entry_name;">\n"] in
                          String.concat "" [s1;s2;s3;">";s4;s5] ;
       end ;

class header =
        object(self)
                value mutable entry_list = [] ;

                method add_entry (e:header_entry) = entry_list := [e] @ entry_list ;

                method rem_entry (e:header_entry) (ls: list header_entry) = match ls with
                   [ [] -> raise No_header_entry_present
                   | [x::l] -> if (x#getname() = e#getname()) then l
                               else [x :: (self#rem_entry e l)]
                   ];

                method get_entrylist () = entry_list ; 

                method to_string () = 
                        let ls = ref entry_list in
                        let s = ref "" in
                        let s1 = "<env:Header> \n " in
                        let s2 = "</env:Header> \n " in
                        let _ = while (List.length ls.val > 0) do
                                {
                                        let e = List.hd ls.val in
                                        s.val := String.concat "" [s.val;e#to_string()] ;
                                        ls.val := List.tl ls.val 
                                } in
                                String.concat "" [s1;s.val;s2] ;

        end ;


class body_entry =
        object (self)
                
                value mutable entry_name = "" ;

                value mutable v = sp_of_struct(make_struct());

                value mutable sch = def ;
                
                value attr_handl = new attributehandler ;

                method setname n = entry_name := n ;

                method getname () = entry_name ;

                method set_sch (s: Pxp_document.node 'a) = sch := s ;
                
                method get_sch () = sch ;

                method setval vv = v := vv ;

                method getval () = v ;
                
                method set_attr q str = match (q,str) with
                                    [ (_,"") -> raise Bad_argument
                                    | (QName n,v) -> attr_handl#set_attr (QName n) v 
                                    ];

                method get_attr q = match q with
                   [QName n -> attr_handl#get_attr (QName n)] ;

                method rem_attr q = match q with
                   [QName n -> attr_handl#rem_attr (QName n)] ;

                method declareNS prefix uri = match (prefix,uri) with
                                    [ ("","") -> ()
                                    | (p,u) -> attr_handl#declareNS p u 
                                    ];

                method to_string () = 
                        let s1 = String.concat "" ["<";entry_name] in
                        let s2 = attr_handl#ns_to_string() in
                        let s3 = attr_handl#att_to_string() in
                        let s4 = SoapTypes.print_value v in
                        let s5 = String.concat "" ["</";entry_name;">\n"] in
                          String.concat "" [s1;s2;s3;">";s4;s5] ;
       end ;

class body = 
        object(self)
        
                value mutable entry_list = [] ;

                method add_entry (e:body_entry) = entry_list := [e] @ entry_list ;

                method rem_entry (e:body_entry) (ls:list body_entry) = match ls with
                    [ [] -> raise No_body_entry_present
                    | [x::l] -> if (x#getname() = e#getname()) then l
                                else [x :: (self#rem_entry e l)]
                    ];

                method get_entrylist () = entry_list ;

                method to_string () = 
                        let ls = ref entry_list in
                        let s = ref "" in
                        let s1 = "<env:Body> \n " in
                        let s2 = "</env:Body> \n " in
                        let _ = while (List.length ls.val > 0) do
                                {
                                        let e = List.hd ls.val in
                                        s.val := String.concat "" [s.val;e#to_string()] ;
                                        ls.val := List.tl ls.val 
                                } in
                                String.concat "" [s1;s.val;s2] ;

        end ;

class envelope =
        object (self)
                value header = ref (Some (new header));
                value body = ref (new body);
                value env_entries = ref [" "] ;
                value attr_handl = new attributehandler ;
                
                method init () = do
                             {
                              self#declareNS soap_env_prefix soap_env_uri ;
                              self#declareNS schema_xsi_prefix schema_xsi_uri ;
                              self#declareNS schema_xsd_prefix schema_xsd_uri ;
                             } ;
                
                method set_attr q str = match (q,str) with
                                    [ (_,"") -> raise Bad_argument
                                    | (QName n,v) -> attr_handl#set_attr (QName n) v 
                                    ];

                method get_attr q = match q with
                   [QName n -> attr_handl#get_attr (QName n)] ;

                method rem_attr q = match q with 
                   [QName n -> attr_handl#rem_attr (QName n)] ;

                method declareNS prefix uri = match (prefix,uri) with
                                    [ ("","") -> ()
                                    | (p,u) -> attr_handl#declareNS p u 
                                    ];
                                    
                method set_header head = do {\:= header head} ;

                method get_header () = match (\!(header)) with
                     [ None -> raise No_header_present
                     | Some h -> h
                     ];

                method set_body b = do {\:= body b} ;

                method get_body () = (\!(body)) ;

                method set_env_entr (ls:list string) = do {\:= env_entries ls} ;

                method get_env_entr () = (\!(env_entries)) ; 

                method private is_null h = match h with
                  [ None -> True
                  | Some hh -> False
                  ];

                method conv_string () = 
                        let s1 = attr_handl#ns_to_string() in
                        let s2 = "<env:Envelope \n " in
                        let s3 = if (self#is_null (\!(header))) then ""
                                 else (self#get_header())#to_string() in
                        let s4 = (\!(body))#to_string() in
                        let s5 = "/env:Envelope >\n" in
                        String.concat "" [s2;s1;">";s3;s4;s5] ;
       end ;

class fault = 
        object
                value faultCode = ref "" ;
                value faultString = ref "" ;
                value faultActorUri = ref "" ;
                value detail_entries = ref [""];
                value fault_entries = ref [""];
                value attr_handl = new attributehandler ;

                method set_attr q str = match (q,str) with
                                    [ (_,"") -> raise Bad_argument
                                    | (QName n,v) -> attr_handl#set_attr (QName n) v 
                                    ];

                method get_attr q = match q with
                   [QName n -> attr_handl#get_attr (QName n)] ;

                method rem_attr q = match q with 
                   [QName n -> attr_handl#rem_attr (QName n)] ;

                method declareNS prefix uri = match (prefix,uri) with
                                    [ ("","") -> ()
                                    | (p,u) -> attr_handl#declareNS p u 
                                    ];

                method set_fault_code f = do {\:= faultCode (f)} ;

                method get_fault_code () = (\!(faultCode)) ;

                method set_fault_string str = do {\:= faultString (str)} ;

                method get_fault_string () = (\!(faultString)) ;

                method set_faultactor_uri a = do {\:= faultActorUri a} ;

                method get_faultactor_uri () = (\!(faultActorUri)) ;

                method set_fault_entr (ls:list string) = do {\:= fault_entries ls} ;

                method get_fault_entr () = (\!(fault_entries)) ; 

                method to_string () = "E" ;
       end ;

value handle_header_entry h_en r = let nname = (nodename r) in
                  let dat = r#data in
                  h_en#set_attr (QName nname) dat ;
                  
value handle_body_entry b_en r = let nname = (nodename r) in
                  let dat = r#data in
                  b_en#set_attr (QName nname) dat ;
                  
value handle_env env r = let nname = (nodename r) in
                  let len = String.length nname in
                  env#declareNS (String.sub nname 5 (len-6)) (r#data) ;
                  
(* We make an assumption that the type information for the data part of the XML
document is present in the header itself. So, we obtain that information and
parse the document accordingly. *)

value handle_head head ls = match ls with
   [ [] -> ()
   | [r::l] ->
          let l1 = head#get_entrylist() in
          set_entries l1 ls where
          rec set_entries ls1 ls2 = match (ls1,ls2) with
          [ ([],[]) -> ()
          | ([],_) -> raise Bad_argument
          | (_,[]) -> raise Bad_argument
          | ([y1::ll1],[y2::ll2]) -> 
                      let sch = y1#get_sch() in
                      let t = y1#get_attr (QName "type") in
                      let v = parse_value sch y2 t in
                      let _ = y1#setval v in
                      set_entries ll1 ll2
          ]
   ];

value handle_body body ls = match ls with
   [ [] -> ()
   | [r::l] ->
          let l1 = body#get_entrylist() in
          set_entries l1 ls where
          rec set_entries ls1 ls2 = match (ls1,ls2) with
          [ ([],[]) -> ()
          | ([],_) -> raise Bad_argument
          | (_,[]) -> raise Bad_argument
          | ([y1::ll1],[y2::ll2]) -> 
                      let sch = y1#get_sch() in
                      let t = y1#get_attr (QName "type") in
                      let v = parse_value sch y2 t in
                      let _ = y1#setval v in
                      set_entries ll1 ll2
          ]
   ];

value handle_head_struct head r = 
          let h_en = new header_entry in
          let ls = r#attributes_as_nodes in
          let _ = h_en#setname (nodename r) in
          let _ = h_en#set_sch r in
          let ls = zip (handle_header_entry h_en) ls in
          head#add_entry h_en;

value handle_body_struct body r = 
          let b_en = new body_entry in
          let ls = r#attributes_as_nodes in
          let _ = b_en#setname (nodename r) in
          let _ = b_en#set_sch r in
          let ls = zip (handle_body_entry b_en) ls in
          body#add_entry b_en;

value get_obj (obj:envelope) s = match s with
   [ "" -> raise Bad_argument
   | s -> let doc =  parse_wfdocument_entity default_config (from_string s) default_spec in
          let r = doc#root in
          let ls = r#sub_nodes in
          let is_h = (nodename (List.hd ls) = "env:Header" ) in
          let b = obj#get_body() in
          let _ = if (not is_h) then 
                                       do {
                                            let br = List.hd ls in
                                            let _ =  handle_body b br#sub_nodes in
                                              () ;
                                            obj#set_body b
                                          }
                  else do
                       {
                               let hr = List.hd (List.tl ls) in
                               let h = obj#get_header () in
                               let _ = handle_head h hr#sub_nodes in 
                                 () ;
                               let br = List.hd (List.tl (List.tl ls)) in
                               let _ = handle_body b br#sub_nodes in
                                 () ;
                               obj#set_header (Some h) ;
                               obj#set_body b
                       }  in
          obj
   ];

value get_obj_struct sch = match sch with
   [ "" -> raise Bad_argument
   | s -> let doc =  parse_wfdocument_entity default_config (from_string s) default_spec in
          let obj = new envelope in
          let h = new header in
          let b = new body in
          let r = doc#root in
          let ls = r#sub_nodes in
          let _ = zip (handle_env obj) (r#attributes_as_nodes) in
          let is_h = (nodename (List.hd ls) = "Header" ) in
          let _ = if (not is_h) then 
                                       do {
                                            obj#set_header None ;
                                            let br = List.hd ls in
                                            let _ =  zip (handle_body_struct b) (br#sub_nodes) in
                                              () ;
                                            obj#set_body b
                                          }
                  else do
                       {
                               let hr = List.hd (List.tl ls) in
                               let _ = zip (handle_head_struct h) (hr#sub_nodes) in 
                                 () ;
                               let br = List.hd (List.tl (List.tl ls)) in
                               let _ = zip (handle_body_struct b) (br#sub_nodes) in
                                 () ;
                               obj#set_header (Some h) ;
                               obj#set_body b
                       }  in
          obj
   ];

value get_string e = (e#conv_string()) ;
