(* Copyright (c) 2002 INRIA *)

(** This module gives a description of the OCAML structure, which represents
a standard soap message *)

open Pxp_yacc ;
open Pxp_document ;
open Pxp_types ;

(** The attributes are named under the [qname] type *)
type qname = [QName of string] ;

(** All the attributes of the [attributehandler] are stored into a
structure of type [table]*)
type table = Hashtbl.t qname string ;

(** [Bad_argument] is raised when there is an erroneous parameter passing
going on in the program *)
exception Bad_argument ;

(** [No_header_present] is raised when the header field of a soap message is
being accessed and the message does not have any *)
exception No_header_present ;

(** This class handles the various attributes and namespaces of different
elements *)
class type attributehandler = 
        object 
                value attributes : table ;
                (** The storehouse of all the attributes present *)

                value namespace_list : ref (list (string * string)) ;
                (** The storehouse of all the declared namespaces *)

                method set_attr : qname -> string -> unit;
                (** Adds/changes an attribute *)

                method get_attr : qname -> string ;
                (** Returns the value of an attribute *)

                method rem_attr : qname -> unit;
                (** Removes the attribute from the [attributes] structure *)

                method declareNS : string -> string -> unit;
                (** Declares a new namespace in the element *)

                method ns_to_string : unit -> string ;
                (** Prints all the namespaces into a string *)

                method att_to_string : unit -> string ;
                (** Prints all the attributes into a string *)

                method to_string : unit -> string;
                (** Prints all the attributes and namespaces into a string *)
        end ;


(** This class encapsulates the entries of the header class *)
class type header_entry = 
        object
                value entry_name : string ;
                (** This is the name of the corresponding element *)

                value v : SoapTypes.ty ;
                (** This will be the of the type SoapTypes.ty *)

                value sch : Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a));
                (** This will be the schema node of the type of value that the
                entry contains *)
                
                value attr_handl : attributehandler;
                (** This is the handler of all the attributes and namespaces of
                the [header] object *)
                
                method setname : string -> unit ;
                (** Assigns the element name to the [header_entry] object *)

                method getname : unit -> string ;
                (** Returns the corresponding element name of the [header_entry]
                object *)

                method setval : SoapTypes.ty -> unit ;
                (** Assigns the element value to the [header_entry] object *)

                method getval : unit -> SoapTypes.ty ;
                (** Returns the corresponding element value of the [header_entry]
                object *)

                method set_sch :  Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a)) -> unit ;
                (** Sets the [sch] field of the object *)

                method get_sch : unit ->  Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a)) ;
                (** Returns the [sch] field of the object *)

                method set_attr : qname -> string -> unit;
                (** Adds/changes an attribute *)

                method get_attr : qname -> string;
                (** Returns the value of an  attribute *)

                method rem_attr : qname -> unit;
                (** Removes the attribute from the [header] structure *)

                method declareNS : string -> string -> unit;
                (** Declares a namespace in the [header] structure *)

                method to_string : unit -> string ;
                (** Prints the header portion of the soap message into a string
                *)
       end ;


(** This class encapsulates the header portion of the soap message *)
class type header =
        object 
                value entry_list : list header_entry;
                (** This is the storehouse of all the entries of the [header]
                structure *)

                method add_entry : header_entry -> unit ;
                (** Adds a new entry to the [header] structure *)

                method rem_entry : header_entry -> list header_entry -> list header_entry ;
                (** Removes a [header_entry] from a list *)

                method get_entrylist : unit -> list header_entry ;
                (** Returns the list of header entries *)
                
                method to_string : unit -> string ;
                (** Prints the header portion of the soap message into a string
                *)
       end ;


(** This class encapsulates the entries of the body class *)
class type body_entry = 
        object
                value entry_name : string ;
                (** This is the name of the corresponding element *)

                value v : SoapTypes.ty ;
                (* This will be the of the type SoapTypes.ty *)

                value sch :  Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a)) ;
                (** This will be the schema node of the type of value that the
                entry contains *)

                value attr_handl : attributehandler;
                (** This is the handler of all the attributes and namespaces of
                the [body] object *)
                
                method setname : string -> unit ;
                (** Assigns the element name to the [body_entry] object *)

                method getname : unit -> string ;
                (** Returns the corresponding element name of the [body_entry]
                object *)

                method setval : SoapTypes.ty -> unit ;
                (** Assigns the element value to the [body_entry] object *)

                method getval : unit -> SoapTypes.ty ;
                (** Returns the corresponding element value of the [body_entry]
                object *)

                method set_sch :  Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a)) -> unit ;
                (** Sets the [sch] field of the object *)

                method get_sch : unit ->  Pxp_document.node (Pxp_document.extension (Pxp_document.node 'a));
                (** Returns the [sch] field of the object *)

                method set_attr : qname -> string -> unit;
                (** Adds/changes an attribute *)

                method get_attr : qname -> string;
                (** Returns the value of an  attribute *)

                method rem_attr : qname -> unit;
                (** Removes the attribute from the [body] structure *)

                method declareNS : string -> string -> unit;
                (** Declares a namespace in the [body] structure *)

                method to_string : unit -> string ;
                (** Prints the body portion of the soap message into a string
                *)
       end ;


(** This class encapsulates the body portion of the soap message *)
class type body =
        object 
                value entry_list : list body_entry;
                (** This is the storehouse of all the entries of the [body]
                structure *)

                method add_entry : body_entry -> unit ;
                (** Adds a new entry to the [body] structure *)

                method rem_entry : body_entry -> list body_entry -> list body_entry ;
                (** Removes a [body_entry] from a list *)

                method get_entrylist : unit -> list body_entry ;
                (** Returns the list of body entries *)
                
                method to_string : unit -> string ;
                (** Prints the body portion of the soap message into a string
                *)
       end ;


(** This class encapsulates the envelope portion of the soap message. In fact, a
soap message is essentially an [envelope] object which itself consists of a
[body], an {i optional } [header] etc. *)
class type envelope =
        object 
                value header : ref (option header);
                (** This is the header part of the soap message. Since it can be
                optional, it is declared as a [option header] *)
                
                value body : ref body;
                (** This is the body part of the soap message *)
                
                value env_entries : ref (list string) ;
                (** This is the storehouse of all the entries of the [envelope]
                structure *)

                value attr_handl : attributehandler ;
                (** This is the handler of all the attributes and namespaces of
                the [envelope] object *)
                
                method init : unit -> unit ;
                (** This function initializes an envelope with the default namespaces
                and uris *)

                method set_attr : qname -> string -> unit;
                (** Adds/changes an attribute *)

                method get_attr : qname -> string;
                (** Returns the value of an  attribute *)

                method rem_attr : qname -> unit;
                (** Removes the attribute from the [envelope] structure *)

                method declareNS : string -> string -> unit;
                (** Declares a namespace in the [envelope] structure *)

                method set_header : option header -> unit;
                (** Sets the [header] of the message *)

                method get_header : unit -> header ;
                (** Returns the [header] of the message *)

                method set_body : body -> unit;
                (** Sets the [body] of the message *)

                method get_body : unit -> body;
                (** Returns the body of the message *)

                method set_env_entr : list string -> unit;
                (** Adds new entries in the [envelope] structure *)

                method get_env_entr : unit -> list string; 
                (** Returns the envelope entries *)

                method conv_string : unit -> string ;
                (** Prints the envelope portion of the soap message into a string *)
       end ;


(** This class encapsulates the fault portion of the soap message. Actually, it
is a special type of soap body. But because of difference in structure and
meaning, it has been treated separately. *)
class type fault = 
        object
                value faultCode : ref string;
                (** This contains the fault code of the fault message *)
                
                value faultString : ref string ;
                (** This contains the fault string of the fault message *)
                
                value faultActorUri : ref string ;
                (** This contains the URI of the fault causing node *)
                
                value detail_entries : ref (list string);
                (** This is the storehouse of all the detail entries of the
                [fault] structure *)

                value fault_entries : ref (list string);
                (** This is the storehouse of all the fault entries of the
                [fault] structure *)

                value attr_handl : attributehandler ;
                (** This is the handler of all the attributes and namespaces of
                the [fault] object *)

                method set_attr : qname -> string -> unit ;
                (** Adds/changes an attribute *)

                method get_attr : qname -> string ;
                (** Returns the value of an  attribute *)

                method rem_attr : qname -> unit ;
                (** Removes the attribute from the [fault] structure *)

                method declareNS : string -> string -> unit ;
                (** Declares a namespace in the [fault] structure *)

                method set_fault_code : string -> unit ;
                (** Sets the fault code of the [fault] structure *)

                method get_fault_code : unit -> string ;
                (** Returns the fault code of the [fault] structure *)

                method set_fault_string : string -> unit ;
                (** Sets the fault string of the [fault] structure *)

                method get_fault_string : unit -> string ;
                (** Returns the fault string of the [fault] structure *)

                method set_faultactor_uri : string -> unit;
                (** Sets the URI of the fault actor  of the [fault] *)

                method get_faultactor_uri : unit -> string ;
                (** Returns the URI of the fault actor  of the [fault] *)

                method set_fault_entr : list string -> unit ;
                (** Adds new entries in the [fault] structure *)

                method get_fault_entr : unit -> list string ; 
                (** Returns the fault entries *)

                method to_string : unit -> string ;
                (** Prints the fault portion of the soap message into a string *)
       end ;

(** Given an XML document and an envelope containing the attributes, this function returns an
[envelope] object *)
value get_obj : envelope -> string -> envelope ;

(** Given an XML document containing a soap message, this function returns an
[envelope] object *)
value get_obj_struct : string -> envelope ;

(** Given an [envelope] object, this function returns the corresponding XML
document *)
value get_string : envelope -> string ;
